@@ -1,6 +1,6 @@
 /* 
    neon SSL/TLS support using OpenSSL
-   Copyright (C) 2002-2011, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2002-2021, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -68,6 +68,8 @@ typedef const unsigned char ne_d2i_uchar;
 #endif
 
 #if OPENSSL_VERSION_NUMBER < 0x10100000L
+#define X509_get0_notBefore X509_get_notBefore
+#define X509_get0_notAfter X509_get_notAfter
 #define X509_up_ref(x) x->references++
 #define EVP_PKEY_up_ref(x) x->references++
 #define EVP_MD_CTX_new() ne_calloc(sizeof(EVP_MD_CTX))
@@ -230,10 +232,10 @@ void ne_ssl_cert_validity_time(const ne_ssl_certificate *cert,
                                time_t *from, time_t *until)
 {
     if (from) {
-        *from = asn1time_to_timet(X509_get_notBefore(cert->subject));
+        *from = asn1time_to_timet(X509_get0_notBefore(cert->subject));
     }
     if (until) {
-        *until = asn1time_to_timet(X509_get_notAfter(cert->subject));
+        *until = asn1time_to_timet(X509_get0_notAfter(cert->subject));
     }
 }
 
@@ -368,7 +370,7 @@ static ne_ssl_certificate *populate_cert(ne_ssl_certificate *cert, X509 *x5)
 }
 
 /* OpenSSL cert verification callback.  This is invoked for *each*
- * error which is encoutered whilst verifying the cert chain; multiple
+ * error which is encountered whilst verifying the cert chain; multiple
  * invocations for any particular cert in the chain are possible. */
 static int verify_callback(int ok, X509_STORE_CTX *ctx)
 {
@@ -578,7 +580,7 @@ ne_ssl_context *ne_ssl_context_create(int mode)
         /* enable workarounds for buggy SSL server implementations */
         SSL_CTX_set_options(ctx->ctx, SSL_OP_ALL);
         SSL_CTX_set_verify(ctx->ctx, SSL_VERIFY_PEER, verify_callback);
-#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+#if !defined(LIBRESSL_VERSION_NUMBER) && OPENSSL_VERSION_NUMBER >= 0x10101000L
         SSL_CTX_set_post_handshake_auth(ctx->ctx, 1);
 #endif
     } else if (mode == NE_SSL_CTX_SERVER) {
@@ -1112,10 +1114,40 @@ char *ne_ssl_cert_export(const ne_ssl_certificate *cert)
     return ret;
 }
 
+static const EVP_MD *hash_to_md(unsigned int flags)
+{
+    switch (flags & NE_HASH_ALGMASK) {
+    case NE_HASH_MD5: return EVP_md5();
+    case NE_HASH_SHA256: return EVP_sha256();
+#ifdef HAVE_OPENSSL11
+    case NE_HASH_SHA512: return EVP_sha512();
+    case NE_HASH_SHA512_256: return EVP_sha512_256();
+#endif
+    default: break;
+    }
+    return NULL;
+}
+
 #if SHA_DIGEST_LENGTH != 20
 # error SHA digest length is not 20 bytes
 #endif
 
+char *ne_ssl_cert_hdigest(const ne_ssl_certificate *cert, unsigned int flags)
+{
+    const EVP_MD *md = hash_to_md(flags);
+    unsigned char dig[EVP_MAX_MD_SIZE];
+    unsigned int len;
+
+    if (!md) return NULL;
+
+    if (!X509_digest(cert->subject, md, dig, &len)) {
+        ERR_clear_error();
+        return NULL;
+    }
+
+    return ne__strhash2hex(dig, len, flags);
+}
+
 int ne_ssl_cert_digest(const ne_ssl_certificate *cert, char *digest)
 {
     unsigned char sha1[EVP_MAX_MD_SIZE];
@@ -1137,8 +1169,33 @@ int ne_ssl_cert_digest(const ne_ssl_certificate *cert, char *digest)
     return 0;
 }
 
+char *ne_vstrhash(unsigned int flags, va_list ap)
+{
+    EVP_MD_CTX *ctx;
+    const EVP_MD *md = hash_to_md(flags);
+    unsigned char v[EVP_MAX_MD_SIZE];
+    unsigned int vlen;
+    const char *arg;
+
+    ctx = EVP_MD_CTX_new();
+    if (!ctx) return NULL;
+
+    if (EVP_DigestInit(ctx, md) != 1) {
+        EVP_MD_CTX_free(ctx);
+        return NULL;
+    }
+
+    while ((arg = va_arg(ap, const char *)) != NULL)
+        EVP_DigestUpdate(ctx, arg, strlen(arg));
+
+    EVP_DigestFinal_ex(ctx, v, &vlen);
+    EVP_MD_CTX_free(ctx);
+
+    return ne__strhash2hex(v, vlen, flags);
+}
+
 #if defined(NE_HAVE_TS_SSL) && OPENSSL_VERSION_NUMBER < 0x10100000L
-/* From OpenSSL 1.1.0 locking callbacks are no longer neede. */
+/* From OpenSSL 1.1.0 locking callbacks are no longer needed. */
 #define WITH_OPENSSL_LOCKING (1)
 
 /* Implementation of locking callbacks to make OpenSSL thread-safe.
@@ -1287,72 +1344,3 @@ void ne__ssl_exit(void)
     }
 #endif
 }
-
-struct ne_md5_ctx {
-    EVP_MD_CTX *ctx;
-};
-
-/* Returns zero on succes, non-zero on failure. */
-static int init_md5_ctx(struct ne_md5_ctx *ctx)
-{
-    ctx->ctx = EVP_MD_CTX_new();
-
-    if (EVP_DigestInit_ex(ctx->ctx, EVP_md5(), NULL) != 1) {
-        return 1;
-    }
-
-    return 0;
-}
-
-struct ne_md5_ctx *ne_md5_create_ctx(void)
-{
-    struct ne_md5_ctx *ctx = ne_malloc(sizeof *ctx);
-    
-    if (init_md5_ctx(ctx)) {
-        ne_free(ctx);
-        return NULL;
-    }
-    
-    return ctx;
-}
-
-void ne_md5_process_block(const void *buffer, size_t len,
-                          struct ne_md5_ctx *ctx)
-{
-    EVP_DigestUpdate(ctx->ctx, buffer, len);
-}
-
-void ne_md5_process_bytes(const void *buffer, size_t len,
-                          struct ne_md5_ctx *ctx)
-{
-    EVP_DigestUpdate(ctx->ctx, buffer, len);
-}
-
-void *ne_md5_finish_ctx(struct ne_md5_ctx *ctx, void *resbuf)
-{
-    EVP_DigestFinal(ctx->ctx, resbuf, NULL);
-    
-    return resbuf;
-}
-
-struct ne_md5_ctx *ne_md5_dup_ctx(struct ne_md5_ctx *ctx)
-{
-    struct ne_md5_ctx *r = ne_md5_create_ctx();
-
-    EVP_MD_CTX_copy_ex(r->ctx, ctx->ctx);
-    
-    return r;
-}
-
-void ne_md5_reset_ctx(struct ne_md5_ctx *ctx)
-{
-    EVP_MD_CTX_reset(ctx->ctx);
-
-    init_md5_ctx(ctx);    
-}
-    
-void ne_md5_destroy_ctx(struct ne_md5_ctx *ctx)
-{
-    EVP_MD_CTX_free(ctx->ctx);
-    ne_free(ctx);
-}
